from dataclasses import fields
import logging

import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.abstract import AbstractCrawler
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from milla_tdf.data_model.stock_price_data import StockPriceMetaData
from milla_tdf.storage.abstract import AbstractStorage
import pyarrow.csv as pv
import pyarrow.parquet as pq
import pyarrow as pa
from os import path

class LocalStorage(AbstractStorage):
    data_folder: str

    @classmethod
    def load_config(cls, data_folder):
        cls.data_folder = data_folder
        TradingFrameWorkConfig.selenium_driver_path = f'{data_folder}/drivers/chromedriver'
        logging.debug(
            f'{cls.__name__} -> loadConfig to folder {cls.data_folder}')

    @classmethod
    def load_data(cls, data_model: AbstractDataModel, meta_data_model:AbstractMetaData =None, crawler: AbstractCrawler = None, **kwargs) ->pd.DataFrame:
        logging.debug(
            f'{cls.__name__} -> loadData data_model {data_model.__name__} ')

        file_path = f'{cls.data_folder}/{data_model.__name__}.parquet'

        if(path.exists(file_path) == True):
            ### load metadata first
            meta_data = cls.load_metadata(file_path=file_path,data_model=AbstractDataModel,meta_data_model=meta_data_model)
            
            local_data = pd.read_parquet(file_path,engine='pyarrow')
            
            crawled_data_df, new_metadata = cls.build_crawl_data_df(data_model=data_model,crawler=crawler,meta_data=meta_data,**kwargs)

            if(meta_data.to_json()== meta_data.to_json() and crawled_data_df.empty == True):
                return crawler.filter_selected_df(local_data, **kwargs)

            union_df =  pd.concat([crawled_data_df,local_data])
            
            model_keys = [field.default_factory() for field in fields(meta_data_model) if field.name == 'keys'][0]

            if(model_keys != []):
                logging.debug(f'''de duplicate keys {model_keys}''')
                union_df = union_df.drop_duplicates(subset=model_keys)

            cls.write_parquet(union_df,new_metadata,data_model)

            return crawler.filter_selected_df(union_df, **kwargs)
        else:
            crawled_data_df, new_metadata = cls.build_crawl_data_df(data_model=data_model,crawler=crawler,meta_data=None,**kwargs)
            cls.write_parquet(crawled_data_df,new_metadata,data_model)
            return crawled_data_df
    
    @classmethod
    def write_parquet(cls,df,new_metadata,data_model):
        table = pa.Table.from_pandas(df)

        custom_meta_json = new_metadata.to_json()
        existing_meta = table.schema.metadata
        combined_meta = {
            'json_str'.encode() : custom_meta_json.encode(),
            **existing_meta
        }

        table = table.replace_schema_metadata(combined_meta)

        pq.write_table(table, f'{cls.data_folder}/{data_model.__name__}.parquet')
        
        logging.info(f'write {data_model.__name__} -> metadata {new_metadata}')
    
    @classmethod
    def build_crawl_data_df(cls,data_model: AbstractDataModel,crawler:AbstractCrawler,meta_data:AbstractMetaData, **kwargs)-> tuple[pd.DataFrame,AbstractMetaData]:
        if(crawler != None):
            data = crawler.crawl_data(data_model=data_model,meta_data=meta_data,**kwargs)
            metadata =  crawler.build_meta_data(meta_data,**kwargs)
            
            result_df =  pd.DataFrame(data)
            
            if(result_df.empty == False):
                cast_string_fields = {}
                for x in fields(data_model):
                    if(x.type == str):
                        cast_string_fields[x.name] = 'string'

                if(cast_string_fields.__len__() != 0):
                    result_df = result_df.astype(cast_string_fields)
            
            return result_df,metadata
        else:
            return None,None

    @classmethod
    def load_metadata(cls,file_path:str,meta_data_model:AbstractMetaData,  **kwargs) -> AbstractMetaData:
        try:
            
            raw_metadata:dict = pq.read_table(file_path).schema.metadata[b'json_str']
            return meta_data_model.from_json(raw_metadata.decode('UTF-8')) 
        except KeyError:
            return None

    @classmethod
    def save_data(cls, data_model: AbstractDataModel, df: pd.DataFrame):
        logging.debug(
            f'{cls.__name__} -> saveData data_model {data_model.__name__}')



from dataclasses import dataclass, field

from dataclasses_json import dataclass_json

from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from datetime import datetime

@dataclass
class StockPriceMetaData(AbstractMetaData):
    symbols: list[str]
    updated: list[str]
    keys:list[str] = field(default_factory=lambda: ['Symbol','Time'])

@dataclass
class StockPriceData(AbstractDataModel):
    Symbol:str
    Time: datetime
    Open: float
    High: float
    Low: float
    Close: float
    Volume: float



from dataclasses import dataclass, field

from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData

@dataclass
class ComFinancialMetaData(AbstractMetaData):
    symbols: list[str]
    updated: list[str]
    keys:list[str] = field(default_factory=lambda: ['Symbol','yearReport','lengthReport'])

@dataclass
class ComFinancialData(AbstractDataModel):
    Symbol:str
    yearReport: float
    lengthReport: float
    bsa1: float = field(init=False)
    bsa2: float
    bsa5: float
    bsa8: float
    bsa10: float
    bsa15: float
    bsa18: float
    bsa23: float
    bsa24: float
    bsa27: float
    bsa29: float
    bsa43: float
    bsa49: float
    bsa50: float
    bsa53: float
    bsa54: float
    bsa55: float
    bsa56: float
    bsa58: float
    bsa67: float
    bsa71: float
    bsa78: float
    bsa79: float
    bsa80: float
    bsa86: float
    bsa90: float
    bsa96: float
    bsa159: float
    bsa162: float
    bsa173: float
    bsa175: float
    bsa209: float
    cashCycle: float
    cfa21: float
    cfa22: float
    comTypeCode: str
    icbCode: str
    isa22: float
    
    lengthReportCal: float
    organCode: str
    rev: float
    rqq72: float
    ryd3: float
    ryd7: float
    ryd11: float
    ryd14: float
    ryd21: float
    ryd25: float
    ryd26: float
    ryd28: float
    ryd30: float
    ryq1: float
    ryq2: float
    ryq3: float
    ryq6: float
    ryq10: float
    ryq12: float
    ryq14: float
    ryq16: float
    ryq18: float
    ryq20: float
    ryq25: float
    ryq27: float
    ryq29: float
    ryq31: float
    ryq34: float
    ryq39: float
    ryq71: float
    ryq76: float
    ryq77: float
    ryq91: float
    
    yearReportCal: float



from copy import copy
from datetime import datetime
import json
import logging
from typing import Union

import pandas as pd
from milla_tdf.config import TradingFrameWorkConfig
from milla_tdf.crawler.abstract import AbstractCrawler, UpdateModeEnum
from milla_tdf.data_model.abstract import AbstractDataModel, AbstractMetaData
from milla_tdf.data_model.stock_price_data import StockPriceData, StockPriceMetaData


class FiinStockPriceDataCrawler(AbstractCrawler):
    data_model = StockPriceData
    update_mode = UpdateModeEnum.APPEND

    @classmethod
    def crawl_data(cls, meta_data: AbstractMetaData, **kwargs) -> list[StockPriceData]:
        super().crawl_data(data_model=cls.data_model, meta_data=meta_data)
        symbols = kwargs['symbols']

        result: list[StockPriceData] = []

        for symbol in symbols:
            latest_update_date = cls.get_latest_update_date(meta_data, symbol)

            if(latest_update_date != None):
                cls.call_api_and_append_result(
                    symbol, latest_update_date, datetime.now().strftime('%Y-%m-%d'), result)

        return result

    @classmethod
    def call_api_and_append_result(cls, symbol: str, start: str, end: str, result: list[StockPriceData], **kwargs):
        date_range_arr = cls.split_calling_arr(start, end)

        for date_range in date_range_arr:
            logging.info(
                f'''calling data {symbol} from {date_range['start']} -> {date_range['end']}''')

            raw_data = json.loads(cls.call_fiin_api_get(f'{TradingFrameWorkConfig.fiin_technical_end_point}/TradingView/GetStockChartData',
                                                        {'language': 'vi', 'Code': symbol, 'Frequency': 'Daily',
                                                         'From': date_range['start'],
                                                            'To': date_range['end'],
                                                            'Type': 'Stock'
                                                         }
                                                        ))['items']

            if(raw_data != None):
                for data in raw_data:
                    result.append(StockPriceData(symbol, datetime.strptime(
                        data['tradingDate'], '%Y-%m-%dT%H:%M:%S'), data['openPrice'], data['highestPrice'], data['lowestPrice'], data['closePrice'], data['totalMatchVolume']))

    @classmethod
    def split_calling_arr(cls, start: str, end: str):
        start = datetime.strptime(start, '%Y-%m-%d')
        end = datetime.strptime(end, '%Y-%m-%d')

        result = []

        while start < end:
            result.append({'start': start.strftime('%Y-%m-%dT%H:%M:%SZ'), 'end': copy(
                start).replace(year=(start.year) + 1).strftime('%Y-%m-%dT%H:%M:%SZ')})
            start = start.replace(year=(start.year) + 1)

        return result

    @classmethod
    def get_latest_update_date(cls, old_meta_data: StockPriceMetaData, symbol: str, **kwargs) -> Union[str, None]:
        if(old_meta_data == None):
            return TradingFrameWorkConfig.data_start_at
        try:
            updated = old_meta_data.updated[old_meta_data.symbols.index(
                symbol)]

            if(updated == datetime.now().strftime('%Y-%m-%d')):
                return None
            else:
                return updated
        except ValueError:
            pass

        return TradingFrameWorkConfig.data_start_at

    @classmethod
    def build_meta_data(cls, old_meta_data: StockPriceMetaData, **kwargs) -> AbstractMetaData:
        super().build_meta_data(old_meta_data, **kwargs)
        if(old_meta_data == None):
            # build new meta data
            symbols = kwargs['symbols']
            updated = [datetime.now().strftime('%Y-%m-%d') for x in symbols]
            return StockPriceMetaData(symbols=symbols, updated=updated)
        else:
            # update old metadata
            return cls.find_and_update_old_metadata(old_meta_data, **kwargs)

    @classmethod
    def find_and_update_old_metadata(cls, old_meta_data: StockPriceMetaData, **kwargs) -> StockPriceMetaData:
        old_symbols = old_meta_data.symbols
        old_updated = old_meta_data.updated
        updated_symbols = kwargs['symbols']
        new_symbols = []
        update_indexes = []
        for symbol in updated_symbols:
            try:
                update_indexes.append(old_symbols.index(symbol))
            except ValueError:
                new_symbols.append(symbol)

        if(len(update_indexes) != 0):
            for update_index in update_indexes:
                old_updated[update_index] = datetime.now().strftime('%Y-%m-%d')

        if(len(new_symbols) != 0):
            for new_symbol in new_symbols:
                old_symbols.append(new_symbol)
                old_updated.append(datetime.now().strftime('%Y-%m-%d'))

        return StockPriceMetaData(old_symbols, old_updated)

    @classmethod
    def filter_selected_df(cls, df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        symbols = kwargs['symbols']

        return df[df['Symbol'].isin(symbols)]



from dataclasses import dataclass, field

from dataclasses_json import dataclass_json

from milla_tdf.data_model.abstract import AbstractDataModel


@dataclass
class SymbolInfoData(AbstractDataModel):
    Symbol:str
    OrganCode:str
    Board:str
    ComTypeCode:str

@dataclass_json
@dataclass
class SymbolInfoMetaData():
    updated:str
    keys:list[str] = field(default_factory=lambda: ['Symbol'])



from datetime import timedelta
import pandas as pd

from milla_tdf.transform.error_codes import TransformError

def transform_data(stock_price_df:pd.DataFrame,financial_df:pd.DataFrame,lagging_days:int):

    if('yearReport' not in financial_df or 'lengthReport' not in financial_df):
        raise TransformError(f'Not found field yearReport & lengthReport')

    quarter_series = financial_df['yearReport'].astype(str) +'-Q' + financial_df['lengthReport'].astype(str)

    financial_df['apply_date'] = pd.PeriodIndex(quarter_series, freq='Q').to_timestamp() + timedelta(days=lagging_days + 90)

    stock_price_df = stock_price_df.sort_values('Time')
    financial_df = financial_df.sort_values('apply_date')

    result = pd.merge_asof(left=stock_price_df,right=financial_df,by='Symbol',left_on='Time',right_on='apply_date',direction='backward')

    return result
